from Player import Player
import Choice

import os,sys
import curses
import signal
import random
import copy

from urwid_reduct.curses_display import Screen

class Sprite:
    def __init__(self, chars):
	self.chars = chars
	self.height = len(chars)
	self.width = len(chars[0])

SPRITES = {
	"3x2": {
	    "Pharoah" : Sprite( 
		['.".',
		 '|-|']),
	    "Djed/" : Sprite( 
		["  /",
		 "/  "]),
	    "Djed\\" : Sprite( 
		["\  ",
		 "  \\"]),
	    "PyramidNE" : Sprite( 
		["\  ",
		 "#_\\"]),
	    "PyramidSW" : Sprite( 
		["\-#",
		 "  \\"]),
	    "PyramidSE" : Sprite( 
		["#-/",
		 "/  "]),
	    "PyramidNW" : Sprite( 
		["  /",
		 "/-#"]),
	    "Obelisk" : Sprite( 
		[".-.",
		 "`-'"]),
	    "Obelisks" : Sprite( 
		[".=.",
		 "`='"]),
	    },
	"2x2": {
	    "Pharoah" : Sprite( 
		["^^",
		 "()"]),
	    "Djed/" : Sprite( 
		[" /",
		 "/ "]),
	    "Djed\\" : Sprite( 
		["\ ",
		 " \\"]),
	    "PyramidNE" : Sprite( 
		["\ ",
		 "#\\"]),
	    "PyramidSW" : Sprite( 
		["\#",
		 " \\"]),
	    "PyramidSE" : Sprite( 
		["#/",
		 "/ "]),
	    "PyramidNW" : Sprite( 
		[" /",
		 "/# "]),
	    "Obelisk" : Sprite( 
		["--",
		 "--"]),
	    "Obelisks" : Sprite( 
		["==",
		 "=="]),
	    },
	"3x3": {
	    "Pharoah" : Sprite( 
		['.".',
		 "|^|",
		 "|-|"]),
	    "Djed/" : Sprite( 
		["  /",
		 " / ",
		 "/  "]),
	    "Djed\\" : Sprite( 
		["\  ",
		 " \ ",
		 "  \\"]),
	    "PyramidNE" : Sprite( 
		["\  ",
		 "#\ ", 
		 "##\\"]),
	    "PyramidSW" : Sprite( 
		["\##",
		 " \#",
		 "  \\"]),
	    "PyramidSE" : Sprite( 
		["##/",
		 "#/ ",
		 "/  "]),
	    "PyramidNW" : Sprite( 
		["  /",
		 " /#",
		 "/##"]),
	    "Obelisk" : Sprite( 
		[".-.",
		 "| |",
		 "`-'"]),
	    "Obelisks" : Sprite( 
		[".=.",
		 "| |",
		 "`='"]),
	    },
	"1x1" : {
	    "Pharoah" : Sprite(['K']),
	    "Obelisks" : Sprite(['O']),
	    "Obelisk" : Sprite(['U']),
	    "Djed/" : Sprite(['/']),
	    "Djed\\" : Sprite(['\\']),
	    "PyramidNW" : Sprite(['d']),
	    "PyramidNE" : Sprite(['b']),
	    "PyramidSW" : Sprite(['q']),
	    "PyramidSE" : Sprite(['p']),
	    },
	}

def monoSprite(piece, player):
    char = { "Pharoah" : 'k',
	    "Obelisks" : 'o',
	    "Obelisk" : 'u',
	    "Djed/" : 'j',
	    "Djed\\" : 'l',
	    "PyramidNW" : 'd',
	    "PyramidNE" : 'b',
	    "PyramidSW" : 'q',
	    "PyramidSE" : 'p'
	    }[piece]
    if player == 0:
	return Sprite([char])
    else:
	return Sprite([char.upper()])

class CursesUI(Player):
    def __init__(self, socket, isServer, gameOpts, opts):
	Player.__init__(self, socket, isServer, gameOpts, opts)
	self.selectedSquare = [self.state.width/2,self.state.height/2]
	self.selectedPieceSquare = None

	if opts.get("monochrome", False):
	    self.mono = True
	    self.origSqWidth = opts.get("sqWidth", 1)
	    self.origSqHeight = opts.get("sqHeight", 1)
	    self.sprites = None
	    self.sprHeight = 1
	    self.sprWidth = 1
	else:
	    self.mono = False
	    self.origSqWidth = opts.get("sqWidth", 5)
	    self.origSqHeight = opts.get("sqHeight", 3)
	    self.sprites = SPRITES[opts.get("spriteset", "3x2")]
	    self.sprHeight = self.sprites["Pharoah"].height
	    self.sprWidth = self.sprites["Pharoah"].width

    def interact(self):
	self.urwidscr = Screen()
	self.urwidscr.run_wrapper(self.wrapped_interact)

    def wrapped_interact(self):
	self.scr = self.urwidscr.s

	for bg in range(8):
	    for fg in range(8):
		try:
		    curses.init_pair(1+8*bg+fg, fg, bg)
		except:
		    pass
	curses.curs_set(2)

	self.setGeometry()

	self.blankScreen()

	self._interact()

    def setGeometry(self):
	self.height, self.width = self.scr.getmaxyx()

	self.sqHeight = self.origSqHeight
	self.sqWidth = self.origSqWidth

	while self.height < self.sqHeight * (self.state.height + 2) + 3:
	    self.sqHeight -= 1
	while self.width < self.sqWidth * (self.state.width + 2) + 6:
	    self.sqWidth -= 1

	if not (self.sqHeight >= self.sprHeight and self.sqWidth >= self.sprWidth):
	    raise "Terminal too small"

	self.geomBoardStart = (1 + self.sqWidth, 1 + self.sqHeight)
		
	self.turnCol = self.geomBoardStart[0] +\
		self.sqWidth * (self.state.width + 1) + 4
	self.scoresCol = self.turnCol + 2

	self.textRow = self.sqHeight * (self.state.height + 3)

    def blankScreen(self):
	for i in range(self.width-1):
	    for j in range(self.height):
		self.scr.addch(j, i, ' ', self.colourPair(curses.COLOR_WHITE))

    def handleSigwinch(self):
	curses.endwin()
	curses.initscr()
	self.setGeometry()
	self.blankScreen()
	self.drawAll()

    def colourPair(self, fg=curses.COLOR_BLACK, bg=curses.COLOR_BLACK):
	return curses.color_pair(1+8*bg + fg)

    def turnRow(self, pl):
	if pl == 0: return self.coordsOfSquare((0, 0))[1]
	else: return self.coordsOfSquare((0, self.state.height - 1))[1]

    def rotated(self, c):
	# Don't rotate board (we'd have to rotate the pieces too...)
	return c
	#if self.player == 1:
	    #return ((self.state.width - 1) - c[0], (self.state.height - 1) - c[1])
	#else:
	    #return c


    # to make things slightly less ugly, we work with 'coords', which are x,y
    # pairs corresponding to the y,x pairs curses wants.
    def coordsOfSquare(self, sq):
	sq = self.rotated(sq)
	return ( self.geomBoardStart[0] +
		self.sqWidth * sq[0] +
		self.sqWidth/2,
		self.geomBoardStart[1] +
		self.sqHeight * sq[1] + self.sqHeight/2 )
    def addch(self, coord, c, attr=0):
	self.scr.addch(coord[1], coord[0], c, attr)
    def move(self, coord):
	self.scr.move(coord[1], coord[0])

    def refresh(self):
	self.move(self.coordsOfSquare(self.selectedSquare))
	self.scr.refresh()

    def laserAttr(self):
	if self.mono:
	    return self.colourPair(curses.COLOR_WHITE)
	else:
	    return self.colourPair(curses.COLOR_RED)

    def drawEdges(self):
	# axes
	for i in range(self.state.width):
	    for j in [-1, self.state.height-1+1]:
		coords = list( self.coordsOfSquare((i,j)) )
		if j == -1 and self.sqHeight <= 2:
		    # to make room for laser indicator
		    coords[1] -= 1
		self.addch( coords,
			chr(ord('a')+i), self.colourPair(curses.COLOR_WHITE))
	for i in [-1, self.state.width-1+1]:
	    for j in range(self.state.height):
		self.addch( self.coordsOfSquare((i,j)),
			chr(ord('1')+j), self.colourPair(curses.COLOR_WHITE))

	# laser indicators
	x,y = self.coordsOfSquare((0,0))
	self.addch( (x,y-self.sqHeight/2-1), '*', self.laserAttr())
	x,y = self.coordsOfSquare((self.state.width-1,self.state.height-1))
	self.addch( (x,y+self.sqHeight/2+1), '*', self.laserAttr())

    def drawState(self):
	self.state.traceLasers()
	for i in range(self.state.width):
	    for j in range(self.state.height):
		self.drawSquare(self.state, [i, j])
	self.drawTurn(self.state.turn)

    def playerColour(self, player):
	if self.mono: return curses.COLOR_WHITE
	if player == 0: return curses.COLOR_WHITE
	elif player == 1: return curses.COLOR_YELLOW
	else: return curses.COLOR_WHITE

    def drawTurn(self, turn):
	self.addch(
		(self.turnCol, self.turnRow(turn)),
		'*',
		self.colourPair(self.playerColour(turn)) | (turn == self.player and curses.A_BOLD))
	self.addch(
		(self.turnCol, self.turnRow(1-turn)),
		' ',
		self.colourPair(self.playerColour(1-turn)))
	if self.state.lastChoice:
	
	    self.scr.addstr(self.coordsOfSquare((0, self.state.height/2))[1] - self.sqHeight/2,
		    self.turnCol - 4, self.state.lastChoice.niceStr())

    def drawScores(self):
	if self.scores == [0,0]: return
	for pl in [0,1]:
	    self.addch(
		(self.scoresCol, self.turnRow(pl)),
		    str(self.scores[pl]),
		    self.colourPair(curses.COLOR_WHITE))

    def gameEnd(self):
	if self.state.dead[0]:
	    if self.state.dead[1]:
		text = "Draw"
	    else:
		text = "White wins"
	elif self.state.dead[1]:
	    text = "Black wins"
	else:
	    text = "Draw"
	self.getKey(text)

    def getKey(self, prompt=""):
	self.scr.addstr(self.textRow, 0, prompt)
	while True:
	    ks = self.urwidscr.get_input()
	    if ks:
		key = ks[0]
		break
	self.scr.move(self.textRow, 0)
	self.scr.clrtoeol()
	return key

    def drawSquare(self, state, coord):
	assert type(coord) == list
	(i,j) = coord
	boldSprite = False
	bg=curses.COLOR_BLACK
	lasere = state.laserTrace.get(coord, 'e')
	lasern = state.laserTrace.get(coord, 'n')
	laserw = state.laserTrace.get(coord, 'w')
	lasers = state.laserTrace.get(coord, 's')
	owner = state.owner(coord)
	selected = (self.selectedPieceSquare == coord)
	hit = (state.lastHit == coord)
	lastMover = (state.lastMover == coord)
	lastSwapped = (state.lastSwapped == coord)
	piece = state.pieceIn(coord)
	if piece:
	    if self.mono:
		spriteFg = curses.COLOR_WHITE
		sprite = monoSprite(piece.type, piece.player)
	    else:
		spriteFg = self.playerColour(piece.player)
		sprite = self.sprites[piece.type]
	    if (selected or lastMover or lastSwapped):
		boldSprite = True
	else:
	    spriteFg = curses.COLOR_WHITE
	    sprite = None
	spriteAttr = self.colourPair(spriteFg)
	if boldSprite:
	    spriteAttr |= curses.A_BOLD
	ownerAttr = self.colourPair(self.playerColour(owner))
	hitAttr = self.laserAttr()

	scrCoord = self.coordsOfSquare((i,j))

	for si in range(self.sqWidth):
	    for sj in range(self.sqHeight):
		charcoord = ( scrCoord[0] - self.sqWidth/2 + si,
			scrCoord[1] - self.sqHeight/2 + sj )
		char = ' '
		charAttr = spriteAttr
		# Sprite:
		if sprite:
		    spi = si - self.sqWidth/2 + sprite.width/2
		    spj = sj - self.sqHeight/2 + sprite.height/2
		    if ( 0 <= spi < sprite.width and 
			    0 <= spj < sprite.height ):
			char = sprite.chars[spj][spi]
		# Laser:
		if char == ' ':
		    laserCol = self.sqWidth / 2
		    laserVertChar = '|'
		    if self.sprHeight % 2 == 0:
			laserRow = self.sqHeight / 2 - 1
			laserHorizChar = '_'
		    else:
			laserRow = self.sqHeight / 2
			laserHorizChar = '-'
		    laserPlayer = -1
		    if ( lasern > -1 and sj <= laserRow and
			    si == laserCol ):
			char = laserVertChar
			laserPlayer = lasern
		    elif ( lasers > -1 and sj > laserRow and
			    si == laserCol ):
			char = laserVertChar
			laserPlayer = lasers
		    elif ( laserw > -1 and si <= laserCol and
			    sj == laserRow and
			    not (sprite and si == laserCol)
			    ):
			char = laserHorizChar
			laserPlayer = laserw
		    elif ( lasere > -1 and si > laserCol and
			    sj == laserRow ):
			char = laserHorizChar
			laserPlayer = lasere
		    if laserPlayer > -1:
			charAttr = self.laserAttr()
			if laserPlayer == state.turn or laserPlayer == 2:
			    charAttr |= curses.A_BOLD
		# piece highlighting (ugly)
		if False:
		    if char == ' ' and (selected or lastMover or lastSwapped):
			charAttr = spriteAttr
			if si == self.sqWidth/2 - self.sprWidth/2 - 1:
			    char = '>'
			elif si == self.sqWidth/2 - self.sprWidth/2 + self.sprWidth + 1:
			    char = '<'
		# Last hit indication ("blood"):
		if char == ' ' and hit and (si+sj)%2:
		    char = '.'
		    charAttr = hitAttr
		# Square ownership:
		if char == ' ' and owner > -1:
		    char = '.'
		    charAttr = ownerAttr
		self.addch(charcoord, char, charAttr)

    def getChoice(self, pl):
	self.selectedPieceSquare = None
	while True:
	    keys = self.urwidscr.get_input()
	    for key in keys:
		move = [0,0]
		if key in ['6', "right", 'l', 's']:
		    move[0] = 1
		elif key in ['4', "left", 'h', 'h']:
		    move[0] = -1
		elif key in ['2','/','-',"down", 'j']:
		    move[1] = 1
		elif key in ['8','*','+',"up", 'k']:
		    move[1] = -1
		elif key in ['7', "home", 'y', 'g']:
		    move[0] = move[1] = -1
		elif key in ['3', "page down", 'n', 'm']:
		    move[0] = move[1] = 1
		elif key in ['1', "end", 'b', 'b']:
		    move[0] = -1
		    move[1] = 1
		elif key in ['9', "page up", 'u', 'c']:
		    move[0] = 1
		    move[1] = -1
		elif key == "ctrl l":
		    self.drawState()
		    self.scr.redrawwin()
		    self.scr.refresh()
		elif key in ['[', '<', '(', ')', '>', ']']:
		    if self.selectedPieceSquare:
			if key in ['[', '<', '(']: dir = 1
			else: dir = -1
			choice = Choice.Choice(self.selectedPieceSquare, self.selectedPieceSquare, dir)
			if self.state.checkLegalChoice(choice, pl):
			    self.selectedPieceSquare = None
			    return choice

		elif key in ['\n', '\r', "enter", ' ', '5']: 
		    if self.selectedPieceSquare != None:
			choice = Choice.Choice(self.selectedPieceSquare, self.selectedSquare)
			if self.state.checkLegalChoice(choice, pl):
			    if self.state.pieceIn(self.selectedPieceSquare).type == "Obelisks":
				choice = Choice.Choice(self.selectedPieceSquare,
					self.selectedSquare, 0, 
					self.getKey("Unstack? [y/N]") == "y")
			    self.selectedPieceSquare = None
			    return choice
		    piece = self.state.pieceIn(self.selectedSquare)
		    if piece and piece.player == pl:
			if self.selectedPieceSquare == self.selectedSquare:
			    self.selectedPieceSquare = None
			    self.drawSquare(self.state, self.selectedSquare)
			else:
			    oldSelectedPiece = copy.copy(self.selectedPieceSquare)
			    self.selectedPieceSquare = self.selectedSquare[:]

			    if oldSelectedPiece: self.drawSquare(self.state, oldSelectedPiece)
			    self.drawSquare(self.state, self.selectedSquare)

		if move != [0,0]:
		    #if pl == 1:
			# board not flipped for black
			#move[0] = -move[0]
			#move[1] = -move[1]
		    oldSelected = self.selectedSquare[:]
		    self.selectedSquare[0] = (
			    (self.selectedSquare[0] + move[0]) % self.state.width )
		    self.selectedSquare[1] = (
			    (self.selectedSquare[1] + move[1]) % self.state.height )
		    self.drawSquare(self.state, oldSelected)
		    self.drawSquare(self.state, self.selectedSquare)

		self.refresh()
