import random
from copy import copy, deepcopy

from Player import Player
from Choice import Choice
from DumbUI import DumbUI

class Robot(Player):
    shortDesc = "???"

def possCoords(width, height):
    for i in range(width):
	for j in range(height):
	    yield (i,j)
def possChoices(state, pl):
    for frm in possCoords(state.width, state.height):
	piece = state.pieceIn(frm) 
	if piece and piece.player == pl:
	    for dir in [[-1,0],[1,0],[0,-1],[0,1],[-1,-1],[1,-1],[-1,1],[1,1]]:
		to = (frm[0] + dir[0], frm[1] + dir[1])
		for unstack in [0] + (piece.type == "Obelisks" and [1] or []):
		    choice = Choice(frm, to, 0, unstack)
		    if state.checkLegalChoice(choice, pl):
			yield choice
	    if piece.rotatable():
		for rotDir in [-1,1]:
		    choice = Choice(frm, frm, rotDir, 0)
		    if state.checkLegalChoice(choice, pl):
			yield choice


class RandomRobot(DumbUI):
    shortDesc = "plays entirely randomly"
    def getChoice(self, pl):
	return random.choice(list(possChoices(self.state, pl)))

class AlphaBeta(DumbUI):
    shortDesc = "Standard minmax algorithm with alphabeta pruning. Default depth 3; set with 'basic:N'"
    def __init__(self, socket, isServer, gameOpts, opts):
	Player.__init__(self, socket, isServer, gameOpts, opts)
	try:
	    self.depth = int(opts["robotParams"])
	except KeyError, ValueError:
	    self.depth = 3
    def getChoice(self, pl):
	def weight(piece, player):
	    if piece is None: return 0
	    return (-1)**(piece.player == player)*{
		    'Obelisk': 1, 'Obelisks': 1, 'PyramidNE': 2,
		    'PyramidNW': 2, 'PyramidSE': 2, 'PyramidSW': 2,
		    'Djed/': 6, 'Djed\\': 6, 'Pharoah': 1000}[piece.type]

	INFTY = 10000
	self.leavesChecked = 0
	def evalNode(depth, value, turn, alpha, beta):
	    if depth == self.depth or self.state.dead[0] or self.state.dead[1]:
		self.leavesChecked += 1
		return value, []
	    tree = []
	    choices = possChoices(self.state, turn)
	    if depth == 0:
		choices = list(choices)
		random.shuffle(choices)
	    for choice in choices:
		hitPiece, stateUpdates = self.state.update(choice, turn)
		childValueRev, childTree = evalNode(depth+1,
			-(value + weight(hitPiece, turn)),
			1-turn, -beta, -alpha)
		childValue = -childValueRev
		#print depth, choice, childValue
		self.state.undoTurn(stateUpdates)
		if childValue > alpha:
		    alpha = childValue
		    tree = [choice] + childTree
		    if alpha >= beta:
			break
	    return alpha, tree

	val, tree = evalNode(0, 0, pl, -INFTY, INFTY)
	print map(str,tree), val, self.leavesChecked
	return tree[0]

class Improved(DumbUI):
    shortDesc = "Experimental version of 'basic'. Should be faster."
    def __init__(self, socket, isServer, gameOpts, opts):
	Player.__init__(self, socket, isServer, gameOpts, opts)
	try:
	    self.depth = int(opts["robotParams"])
	except KeyError, ValueError:
	    self.depth = 3
	self.goodChoices = {}
    def getChoice(self, pl):
	def weight(piece, player):
	    if piece is None: return 0
	    return (-1)**(piece.player == player)*{
		    'Obelisk': 1, 'Obelisks': 1, 'PyramidNE': 2,
		    'PyramidNW': 2, 'PyramidSE': 2, 'PyramidSW': 2,
		    'Djed/': 6, 'Djed\\': 6, 'Pharoah': 1000}[piece.type]

	INFTY = 10000
	self.leavesChecked = 0
	def evalNode(depth, value, turn, alpha, beta):
	    if depth == self.depth or self.state.dead[0] or self.state.dead[1]:
		self.leavesChecked += 1
		return value, []
	    tree = []
	    bestChoice = None
	    goodChoices = self.goodChoices.get(depth, [])
	    choices = possChoices(self.state, turn)
	    if depth == 0:
		choices = list(choices)
		random.shuffle(choices)
	    def orderedChoices():
		# put "good choices" first
		for choice in goodChoices:
		    if self.state.checkLegalChoice(choice, turn):
			yield choice
		for choice in choices:
		    if choice not in goodChoices:
			yield choice
	    for choice in orderedChoices():
		hitPiece, stateUpdates = self.state.update(choice, turn)
		childValueRev, childTree = evalNode(depth+1,
			-(value + weight(hitPiece, turn)),
			1-turn, -beta, -alpha)
		childValue = -childValueRev
		#print depth, choice, childValue
		self.state.undoTurn(stateUpdates)
		if childValue > alpha:
		    alpha = childValue
		    bestChoice = choice
		    tree = [bestChoice] + childTree
		    if alpha >= beta:
			break
	    if bestChoice and bestChoice not in goodChoices:
		goodChoices.append(bestChoice)
		self.goodChoices[depth] = goodChoices
	    return alpha, tree

	val, tree = evalNode(0, 0, pl, -INFTY, INFTY)
	# On our next turn, we'll try the principal line first:
	for i in range(self.depth):
	    try:
		self.goodChoices[i] = [tree[i+2]]
	    except IndexError:
		self.goodChoices[i] = []
	print map(str,tree), val, self.leavesChecked
	return tree[0]


robots = {
	    "random" : RandomRobot,
	    "basic" : AlphaBeta,
	    "improved" : Improved,
	}
