from Layouts import LAYOUTS, TYPECODES
class GameOpts:
    NETSTR_LEN=1
    def __init__(self, layoutIndex=0):
	assert len(LAYOUTS) <= 10
	layoutIndex = max(0, min( len(LAYOUTS)-1, layoutIndex ))
	self.layoutIndex = layoutIndex
    def netStr(self):
	return str(self.layoutIndex)
    def __str__(self):
	return self.netStr()

def netStrToGameOpts(str):
    return GameOpts(int(str))

class State(object):
    def __init__(self, gameOpts):
	self.simultaneous = False
	self.width = 10
	self.height = 8

	self.board = {}

	layout = LAYOUTS[gameOpts.layoutIndex][1]

	for j, row in enumerate(layout):
	    for i, code in enumerate(row):
		if code != ' ':
		    self.board[(i,j)] = Piece(TYPECODES[code.lower()],
			    code.islower())

	self.turn = 0
	self.dead = [False, False]

	self.lastMover = None
	self.lastSwapped = None
	self.lastHit = None

	self.lastChoice = None

	self.traceLasers()

    def inBounds(self, c):
	        return (0 <= c[0] < self.width and 0 <= c[1] < self.height)

    def owner(self, coord):
	if coord[0] == 0: return 0
	if coord[0] == self.width-1: return 1
	if coord[1] in [0, self.height-1]:
	    if coord[0] == 1: return 1
	    if coord[0] == self.width-2: return 0
	return -1

    def pieceIn(self, coord):
	return self.board.get(tuple(coord), None)

    def removePiece(self, coord):
	del self.board[tuple(coord)]

    def coordWrapped(self, coord):
	return ( coord[0] % self.width,
		coord[1] % self.height)

    # All changes to state are abstracted into these StateUpdate objects, so
    # we can efficiently undo the changes (used in AI routines)
    class StateUpdate:
	def __init__(self): pass
	def undo(self): raise "BUG"
	def update(self, state): raise "BUG"
    class ChangeType(StateUpdate):
	def __init__(self, frm, oldtype, newtype):
	    self.frm, self.oldtype, self.newtype = frm, oldtype, newtype
	def undo(self):
	    return State.ChangeType(self.frm, self.newtype, self.oldtype)
	def update(self, state):
	    state.pieceIn(self.frm).type = self.newtype
    class RemovePiece(StateUpdate):
	def __init__(self, pos, type, pl):
	    self.pos, self.type, self.pl = pos, type, pl
	def undo(self):
	    return State.PlacePiece(self.pos, self.type, self.pl)
	def update(self, state):
	    state.removePiece(self.pos)
    class PlacePiece(StateUpdate):
	def __init__(self, pos, type, pl):
	    self.pos, self.type, self.pl = pos, type, pl
	def undo(self):
	    return State.RemovePiece(self.pos, self.type, self.pl)
	def update(self, state):
	    state.board[tuple(self.pos)] = Piece(self.type, self.pl)
    class RotatePiece(StateUpdate):
	def __init__(self, pos, dir):
	    self.pos, self.dir = pos, dir
	def undo(self):
	    return State.RotatePiece(self.pos, -self.dir)
	def update(self, state):
	    state.pieceIn(self.pos).rotate(self.dir)
    class SetDeadState(StateUpdate):
	def __init__(self, pl, b):
	    self.pl, self.b = pl, b
	def undo(self):
	    return State.SetDeadState(self.pl, not self.b)
	def update(self, state):
	    state.dead[self.pl] = self.b

    def update(self, choice, pl):
	assert self.checkLegalChoice(choice, pl)

	moveUpdates = []

	mover = self.pieceIn(choice.frm)
	target = self.pieceIn(choice.to)

	moving = (choice.frm != choice.to)
	unstacking = choice.unstack
	stacking = ( moving and target and
		target.type == "Obelisk" and mover.type == "Obelisk" )
	swapping = (moving and target and not stacking)

	if unstacking:
	    moveUpdates.append(self.ChangeType(choice.frm, "Obelisks", "Obelisk"))
	    moveUpdates.append(self.PlacePiece(choice.to, "Obelisk", pl))
	elif stacking:
	    moveUpdates.append(self.RemovePiece(choice.frm, "Obelisk", pl))
	    moveUpdates.append(self.ChangeType(choice.to, "Obelisk", "Obelisks"))
	elif moving:
	    moveUpdates.append(self.RemovePiece(choice.frm, mover.type, pl))
	    if swapping:
		moveUpdates.append(self.RemovePiece(choice.to, target.type, target.player))

	    moveUpdates.append(self.PlacePiece(choice.to, mover.type, pl))
	    if swapping:
		moveUpdates.append(self.PlacePiece(choice.frm, target.type, target.player))
	else:
	    moveUpdates.append(self.RotatePiece(choice.frm, choice.rotateDir))

	self.processStateUpdates(moveUpdates)

	self.lastMover = list(choice.to)
	if swapping: self.lastSwapped = list(choice.frm)
	else: self.lastSwapped = None

	laserUpdates = []
	hitPos = self.traceLaser(pl)
	hitPiece = None
	if hitPos:
	    hitPiece=self.pieceIn(hitPos)
	    if hitPiece.type == "Obelisks":
		laserUpdates.append(self.ChangeType(hitPos, "Obelisks", "Obelisk"))
	    else:
		laserUpdates.append(self.RemovePiece(hitPos, hitPiece.type, hitPiece.player))
	    if hitPiece.type == "Pharoah":
		laserUpdates.append(self.SetDeadState(pl, True))
	    self.lastHit = hitPos
	else:
	    self.lastHit = None

	self.processStateUpdates(laserUpdates)

	self.lastChoice = choice

	self.turn = 1 - self.turn

	return (hitPiece, moveUpdates + laserUpdates)

    def processStateUpdates(self, stateUpdates):
	for update in stateUpdates:
	    update.update(self)

    def undoTurn(self, stateUpdates):
	cpy = stateUpdates[:]
	cpy.reverse()
	self.processStateUpdates(map(lambda u: u.undo(), cpy))
	self.turn = 1 - self.turn
	
    def traceLaser(self, pl, recordTrace=False):
	if pl == 0:
	    laserDir = [0,1]
	    laserPos = [0,0]
	else:
	    laserDir = [0,-1]
	    laserPos = [self.width-1,self.height-1]
	
	while ( laserDir != [0,0] and 0 <= laserPos[0] <= self.width
		and 0 <= laserPos[1] <= self.height ):
	    if recordTrace:
		self.laserTrace.set(laserPos, laserDir, False, pl)
	    piece = self.pieceIn(laserPos)
	    if piece:
		laserDir = piece.reflect(laserDir)
		if laserDir == [0,0]:
		    return laserPos
	    if recordTrace:
		self.laserTrace.set(laserPos, laserDir, True, pl)
	    laserPos[0] += laserDir[0]
	    laserPos[1] += laserDir[1]
	return None

    def traceLasers(self):
	self.laserTrace = LaserTrace()
	self.traceLaser(0, True)
	self.traceLaser(1, True)

    def checkLegalChoice(self, choice, pl):
	if not ( 0 <= choice.frm[0] < self.width and
		0 <= choice.to[0] < self.width and
		0 <= choice.frm[1] < self.height and
		0 <= choice.to[1] < self.height ):
	    return False
	
	mover = self.pieceIn(choice.frm)
	target = self.pieceIn(choice.to)
	dir = choice.rotateDir

	if mover is None or mover.player != pl:
	    return False

	if dir != 0:
	    if (mover.rotatable() and dir in [1,-1] and
		    choice.to == choice.frm):
		return True
	    return False

	if target and not (
		( mover.type in ["Djed/", "Djed\\"]
		    and target.swappable() ) or
		( mover.type == "Obelisk" and target.type == "Obelisk" and
		    target.player == pl ) ) :
	    return False

	if choice.unstack and mover.type != "Obelisks":
	    return False

	if self.owner(choice.to) == 1-pl:
	    return False

	for dir in [[-1,0],[1,0],[0,-1],[0,1],[-1,-1],[1,-1],[-1,1],[1,1]]:
	    target = (choice.frm[0] + dir[0], choice.frm[1] + dir[1])
	    if target == choice.to:
		return True
	
	return False


class Piece(object):
    DJEDS = [ "Djed/", "Djed\\" ]
    PYRAMIDS = [ "PyramidNE", "PyramidNW", "PyramidSW", "PyramidSE" ]
    def __init__(self, type, player):
	self.type = type
	self.player = player
    def rotatable(self):
	return self.type in (self.DJEDS + self.PYRAMIDS)
    def swappable(self):
	return self.type in (self.PYRAMIDS + [ "Obelisk", "Obelisks" ])
    def rotate(self, dir):
	for metatype in [self.DJEDS, self.PYRAMIDS]:
	    try:
		self.type = metatype[ (metatype.index(self.type) + dir) %
			len(metatype) ]
		return
	    except ValueError:
		pass
    def reflect(self, laserDir):
	if (self.type == "Djed/"
		or (self.type=="PyramidSE" and laserDir in [[-1,0],[0,-1]])
		or (self.type=="PyramidNW" and laserDir in [[1,0],[0,1]])) :
	    return [-laserDir[1],-laserDir[0]]
	elif (self.type == "Djed\\"
		or (self.type=="PyramidNE" and laserDir in [[-1,0],[0,1]])
		or (self.type=="PyramidSW" and laserDir in [[1,0],[0,-1]])) :
	    return [laserDir[1],laserDir[0]]
	else:
	    # absorb
	    return [0,0]


class LaserTrace:
    def __init__(self):
	self.posdict = {}
    def set(self, pos, dir, out, pl):
	pos = tuple(pos)
	dir = tuple(dir)
	part = {
		((-1,0),False) : 'e',
		((-1,0),True) : 'w',
		((1,0),False) : 'w',
		((1,0),True) : 'e',
		((0,-1),False) : 's',
		((0,-1),True) : 'n',
		((0,1),False) : 'n',
		((0,1),True) : 's'
		}[dir,out]
	if not self.posdict.has_key(pos):
	    self.posdict[pos] = {}
	if self.posdict[pos].get(part, -1) == 1-pl:
	    # Both beams
	    self.posdict[pos][part] = 2
	else:
	    self.posdict[pos][part] = pl
    def get(self, pos, part):
	pos = tuple(pos)
	try:
	    return self.posdict[pos][part]
	except KeyError:
	    return -1
